searchData={"items":[{"type":"module","doc":"An authorization library in Elixir for `Plug` and `Phoenix.LiveView` applications that restricts what resources the current user is allowed to access, and automatically load and assigns resources.\n\n`load_resource/2` and `authorize_resource/2` can be used by themselves, while `load_and_authorize_resource/2` combines them both.\n\nThe plug functions are defined in `Canary.Plugs`\n\nIn order to use `Canary` authorization in standard pages with plug, just `import Canary.Plugs` and use plugs, for example:\n\n```elixir\ndefmodule MyAppWeb.PostController do\n  use MyAppWeb, :controller\n  import Canary.Plugs\n\n  plug :load_and_authorize_resource,\n    model: Post,\n    current_user: :current_user,\n    only: [:show, :edit, :update]\nend\n```\n\nThe LiveView hooks are defined in `Canary.Hooks`\n\nIn order to use `Canary` authorization in LiveView, just `use Canary.Hooks` and mount hooks, for example:\n\n```elixir\ndefmodule MyAppWeb.PostLive do\n  use MyAppWeb, :live_view\n  use Canary.Hooks\n\n  mount_canary :load_and_authorize_resource,\n    on: [:handle_params, :handle_event],\n    current_user: :current_user,\n    model: Post,\n    only: [:show, :edit, :update]\n\nend\n```\n\nThis will attach hooks to the LiveView module with `Phoenix.LiveView.attach_hook/4`.\nIn the example above hooks will be attached to `handle_params` and `handle_event` stages of the LiveView lifecycle.\n\nPlease read the documentation for `Canary.Plugs` and `Canary.Hooks` for more information.","title":"Canary","ref":"Canary.html"},{"type":"module","doc":"Hooks functions for loading and authorizing resources for the LiveView events.\nIf you want to authorize `handle_params` and `handle_event` LiveView callbacks\nyou can use `mount_canary` macro to attach the hooks.\n\nYou can think about the `mount_canary` as something similar to `plug` but for LiveView events.\n\nFor `handle_params` it uses `socket.assigns.live_action` as `:action`.\nFor `handle_event` it uses the event name as `:action`.\n\n> Note that the `event_name` is a string - but in Canary it's converted to an atom for consistency.\n\n`Canary.Hooks` and `Canary.Plugs` are separate modules but they share the same API.\nYou can define plugs for standard pages and hooks for LiveView events with the same opts.\n\nFor the authorization actions, when the `:required` is false (by default it's true) it might be nil.\nThen the `Canada.Can` implementation should be the module name of the model rather than a struct.","title":"Canary.Hooks","ref":"Canary.Hooks.html"},{"type":"module","doc":"```elixir\n  use Canary.Hooks\n\n  mount_canary :load_and_authorize_resource,\n    on: [:handle_params, :handle_event],\n    model: Post,\n    only: [:show, :edit, :update]\n\n  mount_canary :authorize_resource,\n    on: [:handle_event],\n    model: Post,\n    only: [:my_event],\n    required: false\n\n  # ...\n\n  def handle_params(params, _uri, socket) do\n    # resource is already loaded and authorized\n    post = socket.assigns.post\n  end\n\n  def handle_event(\"my_event\", _unsigned_params, socket) do\n    # Only admin is allowed to perform my_event\n  end\n\n  ```\n\n  `lib/abilities/user.ex`:\n\n  ```elixir\n  defimpl Canada.Can, for: User do\n\n    def can?(%User{} = user, :my_event, Post), do: user.role == \"admin\"\n\n    def can?(%User{id: id}, _, %Post{user_id: user_id}), do: id == user_id\n  end\n  ```","title":"Example - Canary.Hooks","ref":"Canary.Hooks.html#module-example"},{"type":"function","doc":"Authorize the `:current_user` for the ginve resource. If the `:current_user` is not authorized it will halt the socket.\n\nFor the authorization check, it uses the `can?/3` function from the `Canada.Can` module -\n\n`can?(subject, action, resource)` where:\n\n1. The subject is the `:current_user` from the socket assigns. The `:current_user` key can be changed in the `opts` or in the `Application.get_env(:canary, :current_user, :current_user)`. By default it's `:current_user`.\n2. The action for `handle_params` is `socket.assigns.live_action`, for `handle_event` it uses the event name.\n3. The resource is the loaded resource from the socket assigns or the model name if the resource is not loaded and not required.\n\nRequired opts:\n\n* `:model` - Specifies the module name of the model to load resources from\n* `:on` - Specifies the LiveView lifecycle stages to attach the hook. Default :handle_params\n\nOptional opts:\n\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n>  For `handle_params` it uses `socket.assigns.live_action` as `:action`.\n>  For `handle_event` it uses the event name as `:action`.\n\n* `:as` - Specifies the `resource_name` to get from assigns\n* `:current_user` - Specifies the key in the socket assigns to get the current user\n* `:required` - Specifies if the resource is required, when it's not assigned in socket it will halt the socket\n* `:unauthorized_handler` - Specify a handler function to be called if the action is unauthorized\n\nExample:\n\n```elixir\n\nmount_canary :authorize_resource,\n  model: Post,\n  only: [:show, :edit, :update]\n  current_user: :current_user\n\nmount_canary :authorize_resource,\n  model: Post,\n  as: :custom_resource_name,\n  except: [:new, :create],\n  unauthorized_handler: {ErrorHandler, :unauthorized_handler}\n\n```","title":"Canary.Hooks.authorize_resource/5","ref":"Canary.Hooks.html#authorize_resource/5"},{"type":"function","doc":"Loads and autorize resource and assigns it to the socket. When resource is required it will\nhalt the socket if the resource is not found. If the user is not authorized it will halt the socket.\n\nIt combines `load_resource` and `authorize_resource` functions.\n\nRequired opts:\n\n* `:model` - Specifies the module name of the model to load resources from\n* `:on` - Specifies the LiveView lifecycle stages to attach the hook. Default :handle_params\n\nOptional opts:\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n>  For `handle_params` it uses `socket.assigns.live_action` as `:action`.\n>  For `handle_event` it uses the event name as `:action`.\n\n* `:as` - Specifies the `resource_name` to use in assigns\n* `:current_user` - Specifies the key in the socket assigns to get the current user\n* `:preload` - Specifies association(s) to preload\n* `:id_name` - Specifies the name of the id in `params`, defaults to \"id\"\n* `:id_field` - Specifies the name of the ID field in the database for searching :id_name value, defaults to \"id\".\n* `:required` - Specifies if the resource is required, when it's not found it will halt the socket, default true\n* `:not_found_handler` - Specify a handler function to be called if the resource is not found\n* `:unauthorized_handler` - Specify a handler function to be called if the action is unauthorized\n\nExample:\n\n```elixir\n\nmount_canary :load_and_authorize_resource,\n  model: Comments,\n  id_name: :post_id,\n  id_field: :post_id,\n  only: [:comments]\n\nmount_canary :load_and_authorize_resource,\n  model: Post,\n  as: :custom_name,\n  except: [:new, :create],\n  preload: [:comments],\n  error_handler: CustomErrorHandler\n```","title":"Canary.Hooks.load_and_authorize_resource/5","ref":"Canary.Hooks.html#load_and_authorize_resource/5"},{"type":"function","doc":"Loads the resource and assigns it to the socket. When resource is required it will\nhalt the socket if the resource is not found.\n\n`load_resource` wrapper for attached hook functions, similar to the `load_resource/2` plug\nbut for LiveView events on `:handle_params` and `:handle_event` stages.\n\nRequired opts:\n\n* `:model` - Specifies the module name of the model to load resources from\n* `:on` - Specifies the LiveView lifecycle stages to attach the hook. Default :handle_params\n\nOptional opts:\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n>  For `handle_params` it uses `socket.assigns.live_action` as `:action`.\n>  For `handle_event` it uses the event name as `:action`.\n\n* `:as` - Specifies the `resource_name` to use in assigns\n* `:preload` - Specifies association(s) to preload\n* `:id_name` - Specifies the name of the id in `params`, defaults to \"id\"\n* `:id_field` - Specifies the name of the ID field in the database for searching :id_name value, defaults to \"id\".\n* `:required` - Specifies if the resource is required, when it's not found it will halt the socket\n* `:not_found_handler` - Specify a handler function to be called if the resource is not found\n\nExample:\n\n```elixir\n\nmount_canary :load_resource,\n  model: Post,\n  only: [:show, :edit, :update],\n  preload: [:comments]\n\nmount_canary :load_resource,\n  on: [:handle_params, :handle_event]\n  model: Post,\n  as: :custom_name,\n  except: [:new, :create],\n  preload: [:comments],\n  not_found_handler: {ErrorHandler, :not_found_handler}\n\n```","title":"Canary.Hooks.load_resource/5","ref":"Canary.Hooks.html#load_resource/5"},{"type":"macro","doc":"Mount canary authorization hooks on the current module.\nIt creates a wrapper function to handle_params and handle_event,\nand attaches the hooks to the Live View.","title":"Canary.Hooks.mount_canary/2","ref":"Canary.Hooks.html#mount_canary/2"},{"type":"macro","doc":"```\n  mount_canary :load_and_authorize_resource,\n    model: Post,\n    only: [:edit: :update]\n  ```","title":"Example - Canary.Hooks.mount_canary/2","ref":"Canary.Hooks.html#mount_canary/2-example"},{"type":"module","doc":"Plug functions for loading and authorizing resources for the current request.\n\nThe plugs all store data in conn.assigns (in Phoenix applications, keys in conn.assigns can be accessed with `@key_name` in templates)\n\nIn order to use the plug functions, you must `import Canary.Plugs`.\n\nYou must also specify the Ecto repo to use in your configuration:\n```\nconfig :canary, repo: Project.Repo\n```\nIf you wish, you may also specify the key where Canary will look for the current user record to authorize against:\n```\nconfig :canary, current_user: :some_current_user\n```\n\nYou can specify a error handler module (in this case, `Helpers`) to be called when an action is unauthorized like so:\n```elixir\nconfig :canary, error_handler: Helpers\n```\n\nModule should implement the `Canary.ErrorHandler` behaviour.\n\nCanary will pass the `conn` to the handler function.","title":"Canary.Plugs","ref":"Canary.Plugs.html"},{"type":"function","doc":"Authorize the current user against the calling controller.\n\nIn order to use this function,\n\n  1) `conn.assigns[Application.get_env(:canary, :current_user, :current_user)]` must be an ecto\n  struct representing the current user\n\n  2) `conn.private` must be a map (this should not be a problem unless you explicitly modified it)\n\nauthorize_controller checks for the name of the current controller in one of the following places\n  1) :phoenix_controller in conn.private\n  2) :canary_controller in conn.assigns\n\nIn case you are not using phoenix, make sure you set the controller name in the conn.assigns\nNote that in case neither of `:phoenix_controller` or `:canary_controller` are found the requested\n  authorization won't necessarily fail, rather it will trigger a `.can?` function with a `nil` controller\n\nIf authorization succeeds, sets `conn.assigns.authorized` to true.\n\nIf authorization fails, sets `conn.assigns.authorized` to false.\n\nOptional opts:\n\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n* `:unauthorized_handler` - Specify a handler function to be called if the action is unauthorized\n\nExamples:\n```\nplug :authorize_controller\n\nplug :authorize_controller, only: [:index, :show]\n\nplug :authorize_controller, except: [:destroy]\n```","title":"Canary.Plugs.authorize_controller/2","ref":"Canary.Plugs.html#authorize_controller/2"},{"type":"function","doc":"Authorize the current user for the given resource.\n\nIn order to use this function,\n\n  1) `conn.assigns[Application.get_env(:canary, :current_user, :current_user)]` must be an ecto\n  struct representing the current user\n\n  2) `conn.private` must be a map (this should not be a problem unless you explicitly modified it)\n\nIf authorization succeeds, sets `conn.assigns.authorized` to true.\n\nIf authorization fails, sets `conn.assigns.authorized` to false.\n\nFor the `:index`, `:new`, and `:create` actions, the resource in the `Canada.Can` implementation\nshould be the module name of the model rather than a struct. A struct should be used instead of\nthe module name only if the `:persisted` key is used and you want to override the default\nauthorization behavior.  This can be useful when dealing with nested resources.\n\nFor example:\n\n  use\n  ```\n  def can?(%User{}, :index, Post), do: true\n  ```\n  instead of\n  ```\n  def can?(%User{}, :index, %Post{}), do: true\n  ```\n\n  or\n\n  use\n  ```\n  def can?(%User{id: user_id}, :index, %Post{user_id: user_id}), do: true\n  ```\n  if you are dealing with a nested resource, such as, \"/post/post_id/comments\"\n\n\n  You can specify additional actions for which Canary will authorize based on the model name, by passing the `non_id_actions` opt to the plug.\n\n  For example,\n  ```elixir\n  plug :authorize_resource, model: Post, non_id_actions: [:find_by_name]\n  ```\n\nRequired opts:\n\n* `:model` - Specifies the module name of the model to authorize access to\n\nOptional opts:\n\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n* `:preload` - Specifies association(s) to preload\n* `:id_name` - Specifies the name of the id in `conn.params`, defaults to \"id\"\n* `:id_field` - Specifies the name of the ID field in the database for searching :id_name value, defaults to \"id\".\n* `:persisted` - Specifies the resource should always be loaded from the database, defaults to false\n* `:unauthorized_handler` - Specify a handler function to be called if the action is unauthorized\n\nExamples:\n```\nplug :authorize_resource, model: Post\n\nplug :authorize_resource, model: User, preload: :posts\n\nplug :authorize_resource, model: User, only: [:index, :show], preload: :posts\n\nplug :load_resource, model: Post, id_name: \"post_id\", only: [:index], persisted: true, preload: :comments\n\nplug :load_resource, model: Post, id_name: \"slug\", id_field: \"slug\", only: [:show], persisted: true\n```","title":"Canary.Plugs.authorize_resource/2","ref":"Canary.Plugs.html#authorize_resource/2"},{"type":"function","doc":"Authorize the given resource and then load it if\nauthorization succeeds.\n\nIf the resource cannot be loaded or authorization\nfails, conn.assigns.resource_name is set to nil.\n\nThe result of the authorization (true/false) is\nassigned to conn.assigns.authorized.\n\nAlso, see the documentation for load_resource/2 and\nauthorize_resource/2.\n\nRequired opts:\n\n* `:model` - Specifies the module name of the model to load resources from\n\nOptional opts:\n\n* `:as` - Specifies the `resource_name` to use\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n* `:preload` - Specifies association(s) to preload\n* `:id_name` - Specifies the name of the id in `conn.params`, defaults to \"id\"\n* `:id_field` - Specifies the name of the ID field in the database for searching :id_name value, defaults to \"id\".\n* `:unauthorized_handler` - Specify a handler function to be called if the action is unauthorized\n* `:not_found_handler` - Specify a handler function to be called if the resource is not found\n\nNote: If both an `:unauthorized_handler` and a `:not_found_handler` are specified for `load_and_authorize_resource`,\nand the request meets the criteria for both, the `:unauthorized_handler` will be called first.\n\nExamples:\n```\nplug :load_and_authorize_resource, model: Post\n\nplug :load_and_authorize_resource, model: User, preload: :posts, as: :the_user\n\nplug :load_and_authorize_resource, model: User, only: [:index, :show], preload: :posts, as: :person\n\nplug :load_and_authorize_resource, model: User, except: [:destroy]\n\nplug :load_and_authorize_resource, model: Post, id_name: \"slug\", id_field: \"slug\", only: [:show], persisted: true\n```","title":"Canary.Plugs.load_and_authorize_resource/2","ref":"Canary.Plugs.html#load_and_authorize_resource/2"},{"type":"function","doc":"Load the given resource.\n\nLoad the resource with id given by `conn.params[\"id\"]` (or `conn.params[opts[:id_name]]` if `opts[:id_name]` is specified)\nand ecto model given by `opts[:model]` into `conn.assigns.resource_name`.\n\n`resource_name` is either inferred from the model name or specified in the plug declaration with the `:as` key.\nTo infer the `resource_name`, the most specific(right most) name in the model's\nmodule name will be used, converted to underscore case.\n\nFor example, `load_resource model: Some.Project.BlogPost` will load the resource into\n`conn.assigns.blog_post`\n\nIf the resource cannot be fetched, `conn.assigns.resource_name` is set\nto nil.\n\nBy default, when the action is `:index`, all records from the specified model will be loaded. This can\nbe overridden to fetch a single record from the database by using the `:persisted` key.\n\nCurrently, `:new` and `:create` actions are ignored, and `conn.assigns.resource_name`\nwill be set to nil for these actions. This can be overridden to fetch a single record from the database\nby using the `:persisted` key.\n\nThe `:persisted` key can override how a resource is loaded and can be useful when dealing\nwith nested resources.\n\nRequired opts:\n\n* `:model` - Specifies the module name of the model to load resources from\n\nOptional opts:\n\n* `:as` - Specifies the `resource_name` to use\n* `:only` - Specifies which actions to authorize\n* `:except` - Specifies which actions for which to skip authorization\n* `:preload` - Specifies association(s) to preload\n* `:id_name` - Specifies the name of the id in `conn.params`, defaults to \"id\"\n* `:id_field` - Specifies the name of the ID field in the database for searching :id_name value, defaults to \"id\".\n* `:persisted` - Specifies the resource should always be loaded from the database, defaults to false\n* `:required` - Same as `:persisted` but with not found handler - even for :index, :new or :create action\n* `:not_found_handler` - Specify a handler function to be called if the resource is not found\n\n\nExamples:\n```\nplug :load_resource, model: Post\n\nplug :load_resource, model: User, preload: :posts, as: :the_user\n\nplug :load_resource, model: User, only: [:index, :show], preload: :posts, as: :person\n\nplug :load_resource, model: User, except: [:destroy]\n\nplug :load_resource, model: Post, id_name: \"post_id\", only: [:new, :create], persisted: true\n\nplug :load_resource, model: Post, id_name: \"slug\", id_field: \"slug\", only: [:show], persisted: true\n```","title":"Canary.Plugs.load_resource/2","ref":"Canary.Plugs.html#load_resource/2"},{"type":"module","doc":"Common utils functions for `Canary.Plugs` and `Canary.Hooks`","title":"Canary.Utils","ref":"Canary.Utils.html"},{"type":"function","doc":"Check if an action is valid based on the options.\n\n    iex> Canary.Utils.action_valid?(:index, only: [:index, :show])\n      true\n\n    iex> Canary.Utils.action_valid?(:index, except: :index)\n      false\n\n    iex> Canary.Utils.action_valid?(:show, except: :index, only: :show)\n      ** (ArgumentError) You can't use both :except and :only options","title":"Canary.Utils.action_valid?/2","ref":"Canary.Utils.html#action_valid?/2"},{"type":"function","doc":"Apply the error handler to the connection or socket","title":"Canary.Utils.apply_error_handler/3","ref":"Canary.Utils.html#apply_error_handler/3"},{"type":"function","doc":"Check if the not_found handler should be applied for given action, assigns and options","title":"Canary.Utils.apply_handle_not_found?/3","ref":"Canary.Utils.html#apply_handle_not_found?/3"},{"type":"function","doc":"Get the resource id from the connection params\n\n    iex> Canary.Utils.get_resource_id(%{\"id\" => \"9\"}, [])\n    \"9\"\n\n    iex> Canary.Utils.get_resource_id(%Plug.Conn{params: %{\"custom_id\" => \"1\"}}, id_name: \"custom_id\")\n    \"1\"\n\n    iex> Canary.Utils.get_resource_id(%{\"user_id\" => \"7\"}, id_name: \"user_id\")\n    \"7\"\n\n    iex> Canary.Utils.get_resource_id(%{\"other_id\" => \"9\"}, id_name: \"id\")\n    nil","title":"Canary.Utils.get_resource_id/2","ref":"Canary.Utils.html#get_resource_id/2"},{"type":"function","doc":"Get the resource name from the options, covert it to atom and pluralize\nit if needed - only for :index action.\n\nIf the `:as` option is provided, it will be used as the resource name.\n\n      iex> Canary.Utils.get_resource_name(:show, model: MyApp.Post)\n      :post\n\n      iex> Canary.Utils.get_resource_name(:index, model: MyApp.Post)\n      :posts\n\n      iex> Canary.Utils.get_resource_name(:index, model: MyApp.Post, as: :my_posts)\n      :my_posts","title":"Canary.Utils.get_resource_name/2","ref":"Canary.Utils.html#get_resource_name/2"},{"type":"function","doc":"Returns the :non_id_actions option if it is present","title":"Canary.Utils.non_id_actions/1","ref":"Canary.Utils.html#non_id_actions/1"},{"type":"function","doc":"","title":"Canary.Utils.persisted?/1","ref":"Canary.Utils.html#persisted?/1"},{"type":"function","doc":"Preload associations if needed","title":"Canary.Utils.preload_if_needed/3","ref":"Canary.Utils.html#preload_if_needed/3"},{"type":"function","doc":"Check if a key is present in a keyword list","title":"Canary.Utils.required?/1","ref":"Canary.Utils.html#required?/1"},{"type":"module","doc":"The fallback Canary handler.\n\nThis module is used primarily as a backwards compatibility for the `:not_found_handler` and `:unauthorized_handler`.\nIt uses old configuration values to determine how to handle the error.\n\nIf you are using `Canary` only with `Plug` based authorization then you can still\nuse the `:not_found_handler` and `:unauthorized_handler` configuration values.\nOtherwise, you should implement the `Canary.ErrorHandler` behaviour in your own module.","title":"Canary.DefaultHandler","ref":"Canary.DefaultHandler.html"},{"type":"function","doc":"The default handler for when a resource is not found.\nFor Plug based authorization it will use the global `:not_found_handler` or return the conn.\n\nFor LiveView base authorization it will halt socket.","title":"Canary.DefaultHandler.not_found_handler/1","ref":"Canary.DefaultHandler.html#not_found_handler/1"},{"type":"function","doc":"The default handler for when a resource is not authorized.\nFor Plug based authorization it will use the global `:unauthorized_handler` or return the conn.\n\nFor LiveView base authorization it will halt socket.","title":"Canary.DefaultHandler.unauthorized_handler/1","ref":"Canary.DefaultHandler.html#unauthorized_handler/1"},{"type":"behaviour","doc":"Specifies the behavior for handling errors in Canary.","title":"Canary.ErrorHandler","ref":"Canary.ErrorHandler.html"},{"type":"callback","doc":"Handles the case where a resource is not found.","title":"Canary.ErrorHandler.not_found_handler/1","ref":"Canary.ErrorHandler.html#c:not_found_handler/1"},{"type":"callback","doc":"Handles the case where a resource is not authorized.","title":"Canary.ErrorHandler.unauthorized_handler/1","ref":"Canary.ErrorHandler.html#c:unauthorized_handler/1"},{"type":"extras","doc":"# Getting Started\n\nThis guide introduces **Canary**, an authorization library for **Elixir** applications using `Plug` and `Phoenix.LiveView`. It restricts resource access based on user permissions and automatically loads and assigns resources.\n\nCanary provides three primary functions to be used as *plugs* or *LiveView hooks* to manage resources:\n\n- `load_resource`\n- `authorize_resource`\n- `load_and_authorize_resource`","title":"Getting Started","ref":"getting-started.html"},{"type":"extras","doc":"","title":"Glossary - Getting Started","ref":"getting-started.html#glossary"},{"type":"extras","doc":"The key name used to fetch the subject from `assigns`. This subject is passed to `Canada.Can` to evaluate permissions. By default, it is `:current_user`.\n\nTo configure this in your module:\n\n```elixir\nconfig :canary, current_user: :user\n```\n\nYou can override this setting per plug/mounted hook by specifying `:current_user`.","title":"Subject - Getting Started","ref":"getting-started.html#subject"},{"type":"extras","doc":"For **Phoenix applications and Plug-based pages**, Canary determines the action automatically from `conn.private.phoenix_action`. In **non-Phoenix applications**, or when overriding Phoenix's default action behavior, set `conn.assigns.canary_action` with an atom specifying the action.\n\nFor **LiveView**:\n\n- In `handle_params`, Canary uses `socket.assigns.live_action`.\n- In `handle_event`, Canary uses the `event_name` (converted from a string to an atom for consistency).\n\nActions can be limited using `:only` or `:except` options; otherwise, they apply to all actions.","title":"Action - Getting Started","ref":"getting-started.html#action"},{"type":"extras","doc":"For `load_resource` and `load_and_authorize_resource`, Canary checks if the resource is already assigned. If not, it fetches the resource from the repository using:\n\n- `:id_name` from `params` (default: `\"id\"`).\n- `:id_field` in the struct (default: `:id`).\n\nBy default, **a resource is required**. That means the resource must be present in `conn.assigns` or `socket.assigns`. It's fetched using the `:model` name, which can be overridden with the `:as` option.\n\nIf it cannot be found, an error is handled. To make it optional, set `:required` to `false`. In this case, the resource module name is used instead of a loaded struct.\n\nYou can also use `:preload` to preload associations. See `Ecto.Query.preload/3` for more details.\n\nFor `authorize_resource`, the resource must be present in `conn.assigns` or `socket.assigns`. By default, it fetches the resource using the `:model` name, which can be overridden with the `:as` option.","title":"Resource - Getting Started","ref":"getting-started.html#resource"},{"type":"extras","doc":"Loads a resource from the database using the specified **Ecto repo** and model. It assigns the result to `assigns. `, where `resource_name` is inferred from the model.","title":"Load Resource - Getting Started","ref":"getting-started.html#load-resource"},{"type":"extras","doc":"Checks if the **subject** can perform a given action on a resource. The result (`true`/`false`) is assigned to `assigns.authorized`. The developer decides how to handle this result.","title":"Authorize Resource - Getting Started","ref":"getting-started.html#authorize-resource"},{"type":"extras","doc":"A combination of **Load Resource** and **Authorize Resource** in a single function.","title":"Load and Authorize Resource - Getting Started","ref":"getting-started.html#load-and-authorize-resource"},{"type":"extras","doc":"To use Canary, you need to configure it in `config/config.exs`. All settings, except for `:repo`, can be overridden when using the plug or hook.","title":"Configuration - Getting Started","ref":"getting-started.html#configuration"},{"type":"extras","doc":"| Name | Description | Example |\n| --- | --- | --- |\n| `:repo` | The Repo module used in your application. | `YourApp.Repo` |\n| `:current_user` | The key name used to fetch the user from assigns. This value will be used as the `subject` for `Canada.Can` to evaluate permissions. Defaults to `:current_user`. | `:current_member` |\n| `:error_handler` | A module that implements the `Canary.ErrorHandler` behavior. It is used to handle `:not_found` and `:unauthorized` errors. Defaults to `Canary.DefaultHandler`. | `YourApp.ErrorHandler` |","title":"Available Configuration Options - Getting Started","ref":"getting-started.html#available-configuration-options"},{"type":"extras","doc":"| Name | Description | Example |\n| --- | --- | --- |\n| `:not_found_handler` | A `{mod, fun}` tuple for handling not found errors. | `{YourApp.ErrorHandler, :handle_not_found}` |\n| `:unauthorized_handler` | A `{mod, fun}` tuple for handling unauthorized errors. | `{YourApp.ErrorHandler, :handle_unauthorized}` |\n\n> #### Info {: .info}\n>\n> The `:error_handler` option should be used instead of separate handlers for `:not_found` and `:unauthorized` errors.\n> Handlers can still be overridden using plug or `mount_canary` options.","title":"Deprecated Options - Getting Started","ref":"getting-started.html#deprecated-options"},{"type":"extras","doc":"```elixir\nconfig :canary,\n  repo: YourApp.Repo,\n  current_user: :current_user,\n  error_handler: YourApp.ErrorHandler\n```","title":"Example Configuration - Getting Started","ref":"getting-started.html#example-configuration"},{"type":"extras","doc":"#### Authorize different subject\nSometimes, you may need to perform authorization for a different subject. You can override `:current_user` by passing options to the plug or hook.\n\n<!-- tabs-open -->","title":"Overriding configuration - Getting Started","ref":"getting-started.html#overriding-configuration"},{"type":"extras","doc":"```elixir\nimport Canary.Plugs\n\nplug :load_and_authorize_resource,\n  model: Team,\n  current_user: :current_member\n```\n\nWith this override, the authorization check will use `conn.assigns.current_member` as the subject.","title":"Conn Plugs - Getting Started","ref":"getting-started.html#conn-plugs"},{"type":"extras","doc":"```elixir\nuse Canary.Hooks\n\nmount_canary :load_and_authorize_resource,\n  on: :handle_event,\n  current_user: :current_member,\n  model: Team\n```\n\nWith this override, the authorization check for the `:handle_event` stage hook will use `socket.assigns.current_member` as the subject.\n<!-- tabs-close -->","title":"LiveView Hooks - Getting Started","ref":"getting-started.html#liveview-hooks"},{"type":"extras","doc":"If you want to override the global Canary error handler, you can override one of the functions: `:not_found_handler` or `:unauthorized_handler`.\n\n<!-- tabs-open -->","title":"Different error handler - Getting Started","ref":"getting-started.html#different-error-handler"},{"type":"extras","doc":"```elixir\nplug :load_and_authorize_resource,\n  model: Team,\n  current_user: :current_member,\n  not_found_handler: {CustomErrorHandler, :custom_not_found_handler},\n  unauthorized_handler: {CustomErrorHandler, :custom_unauthorized_handler}\n```","title":"Conn Plugs - Getting Started","ref":"getting-started.html#conn-plugs"},{"type":"extras","doc":"```elixir\nuse Canary.Hooks\n\nmount_canary :load_and_authorize_resource,\n  model: Team,\n  current_user: :current_member,\n  only: [:special_action]\n  unauthorized_handler: {CustomErrorHandler, :special_unauthorized_handler}\n```\n<!-- tabs-close -->\n\nThe error handler should implement the `Canary.ErrorHandler` behavior.\nRefer to the default implementation in `Canary.DefaultHandler`.","title":"LiveView Hooks - Getting Started","ref":"getting-started.html#liveview-hooks"},{"type":"extras","doc":"Canary Plugs and Hooks use the same configuration options.","title":"Canary options - Getting Started","ref":"getting-started.html#canary-options"},{"type":"extras","doc":"| Name | Description | Example |\n| --- | --- | --- |\n| `:model` | The model module name used in your app. **Required** | `Post` |\n| `:only` | Specifies the actions for which the plug/hook is enabled. | `[:show, :edit, :update]` |\n| `:except` | Specifies the actions for which the plug/hook is disabled. | `[:delete]` |\n| `:current_user` | The key name used to fetch the user from assigns. This value will be used as the `subject` for `Canada.Can` to evaluate permissions. Defaults to `:current_user`. Applies only to `authorize_resource` or `load_and_authorize_resource`. | `:current_member` |\n| `:on` | Specifies the LiveView lifecycle stages where the hook should be attached. Defaults to `:handle_params`. **Available only in Canary.Hooks** | `[:handle_params, :handle_event]` |\n| `:as` | Specifies the key name under which the resource will be stored in assigns. | `:team_post` |\n| `:id_name` | Specifies the name of the ID in params. *Defaults to `\"id\"`*. | `:post_id` |\n| `:id_field` | Specifies the database field name used to search for the `id_name` value. *Defaults to `\"id\"`*. | `:post_id` |\n| `:required` | Determines if the resource is required. If not found, it triggers a not found error. *Defaults to `true`*. | `false` |\n| `:not_found_handler` | A `{mod, fun}` tuple that overrides the default error handler for not found errors. | `{YourApp.ErrorHandler, :custom_handle_not_found}` |\n| `:unauthorized_handler` | A `{mod, fun}` tuple that overrides the default error handler for unauthorized errors. | `{YourApp.ErrorHandler, :custom_handle_unauthorized}` |","title":"Available Options - Getting Started","ref":"getting-started.html#available-options"},{"type":"extras","doc":"| Name | Description | Example |\n| --- | --- | --- |\n| `:non_id_actions` | Additional actions for which Canary will authorize based on the model name. | `[:index, :new, :create]` |\n| `:persisted` | Forces the resource to always be loaded from the database. Defaults to `false`. **Available only in Canary.Plugs** | `true` |","title":"Deprecated Options - Getting Started","ref":"getting-started.html#deprecated-options"},{"type":"extras","doc":"```elixir\n  plug :load_and_authorize_resource,\n    current_user: :current_member,\n    model: Machine,\n    preload: [:plan, :networks, :distribution, :job, ipv4: [:ip_pool], hypervisor: :region]\n\n  plug :load_resource,\n    model: Hypervisor,\n    id_name: \"hypervisor_id\",\n    only: [:new, :create],\n    preload: [:region, :hypervisor_type, machines: [:networks, :plan, :distribution]],\n\n  plug :load_and_authorize_resource,\n    model: Hypervisor,\n    preload: [\n      :region,\n      :hypervisor_type,\n      machines:\n        {Hypervisors.preload_active_machines, [:plan, :distribution, :hypervisor, :networks]}\n    ]\n\n  mount_canary :authorize_resource,\n    on: [:handle_params, :handle_event],\n    current_user: :current_member,\n    model: Machine,\n    only: [:index, :new],\n    required: false\n\n  mount_canary :load_and_authorize_resource,\n    on: [:handle_event],\n    current_user: :current_member,\n    model: Machine,\n    only: [:start, :stop, :restart, :poweroff]\n```","title":"Examples - Getting Started","ref":"getting-started.html#examples"},{"type":"extras","doc":"`Canary.Plugs` and `Canary.Hooks` should work the same way in most cases, providing a unified approach to authorization for both Plug-based controllers and LiveView.\n\n- **Shared Functionality:**\n  Both Plugs and Hooks allow for resource loading and authorization using similar configuration options. This ensures consistency across different parts of your application.\n\n- **Differences:**\n  - `Canary.Plugs` is designed for use in traditional Phoenix controllers and pipelines.\n  - `Canary.Hooks` is specifically built for LiveView and integrates with lifecycle events such as `:handle_params` and `:handle_event`.\n\n- **Configuration Compatibility:**\n  Most options, such as `:model`, `:current_user`, `:only`, `:except`, and error handlers, function identically in both Plugs and Hooks. However, `Canary.Hooks` includes the `:on` option, allowing you to specify which LiveView lifecycle stage the authorization should run on.\n\nBy keeping their behavior aligned, Canary ensures a seamless developer experience, whether you're working with traditional controller-based actions or real-time LiveView interactions.","title":"Plug and Hooks - Getting Started","ref":"getting-started.html#plug-and-hooks"},{"type":"extras","doc":"The `authorize_resource` function checks whether the subject, typically stored in `assigns` under `:current_user`, is authorized to access a given resource. If the `:current_user` is not authorized, it sets `assigns.authorized` to `false` and calls the `handle_unauthorized/1` function from the `:error_handler` module configured in `config.exs` or the `:unauthorized_handler` specified in the options.\n\n#### Authorization Logic\n\nThe authorization check is performed using the `can?/3` function from the `Canada.Can` protocol implemeted for `subject`:\n\n```elixir\ncan?(subject, action, resource)\n```\n\nwhere:\n\n1. **Subject** – The entity being authorized, typically fetched from `assigns.current_user`.\n   - By default, Canary looks for `:current_user`.\n   - This key can be overridden via the `opts` or globally in `Application.get_env(:canary, :current_user, :current_user)`.\n\n2. **Action** – The current action being performed.\n\n3. **Resource** – The resource being accessed.\n   - If the resource is already loaded, it is taken from `assigns`.\n   - If the resource is not loaded and not required, the model name is used instead.\n\n#### Example Usage\n\n```elixir\n# Replace `plug` with `mount_canary` for LiveView Hooks\nplug :authorize_resource,\n  current_user: :current_member,\n  model: Event,\n  as: :public_event\n```\n\nIn this example:\n\n1. The `authorize_resource` function checks whether `:current_member` (instead of the default `:current_user`) is authorized to access the `Event` resource.\n2. The resource is expected to be available in `assigns.public_event`.\n3. If the user is unauthorized, `assigns.authorized` is set to false, and the `unauthorized_handler` is triggered.","title":"Authorize Resource - Getting Started","ref":"getting-started.html#authorize-resource"},{"type":"extras","doc":"The `load_resource` function fetches a resource based on an ID provided in `params` and assigns it to `assigns`. By default, it uses the `\"id\"` key from `params` and retrieves the resource from the database using the `:id` field of the model specified in `opts[:model]`. The loaded resource is stored under `assigns` using a key derived from the model module name.\n\n#### Customizing the Load Behavior\n\nYou can modify the default behavior with the following options:\n\n- **`:id_name`** – Override the default `\"id\"` param key.\n- **`:id_field`** – Change the field used to query the resource in the database.\n- **`:as`** – Override the default `assigns` key where the resource is stored.\n- **`:required`** - When set to `false` it will assign `nil` instad calling the `not_found_handler`.\n\n#### Example Usage\n\n```elixir\n# Replace `plug` with `mount_canary` for LiveView Hooks\nplug :load_resource,\n  model: Event,\n  as: :public_event,\n  id_name: \"uuid\",\n  id_field: :uuid,\n  required: false\n```\n\nIn this example:\n\n1. `load_resource` fetches the `\"uuid\"` from `params`.\n2. It queries `Event` using the `:uuid` field in the database.\n3. The result is assigned to `assigns.public_event`.\n4. If no matching `Event` is found, `assigns.public_event` will be set to `nil`.\n\nTo trigger the `not_found_handler` when the resource is missing, ensure the `:required` flag is **not explicitly set to** `false` (it defaults to `true`).","title":"Load Resource - Getting Started","ref":"getting-started.html#load-resource"},{"type":"extras","doc":"The `load_and_authorize_resource` function combines two operations:\n\n1. **Loading the Resource** – Fetches the resource based on an ID from `params` and assigns it to `assigns`, similar to `load_resource`.\n2. **Authorizing the Resource** – Checks whether the subject (by default, `:current_user`) is authorized to access the resource, using `authorize_resource`.\n\nThis function ensures that resources are both retrieved and access-controlled within a single step.\n\n> #### Error handler order {: .info}\n>\n> If both `:unauthorized_handler` and `:not_found_handler` are specified for `load_and_authorize_resource`, and the request meets the criteria for both, the `:unauthorized_handler` will be called first.","title":"Load and Authorize Resource - Getting Started","ref":"getting-started.html#load-and-authorize-resource"},{"type":"extras","doc":"For actions that do not require loading a specific resource (such as `:index`, `:new`, and `:create`), use `:authorize_resource` instead of `:load_resource` or `:load_and_authorize_resource`.\nEnsure that these functions are limited to actions where resource loading is necessary.\n\nBy default, the `:required` option is set to `true`, meaning that if the resource cannot be found in the repository, the `not_found_handler` will be called.\nSetting `:required` to `false` allows the resource to be assigned as `nil`, in which case the model module name will be used as the resource when calling `can?/3`.","title":"Non-ID Actions - Getting Started","ref":"getting-started.html#non-id-actions"},{"type":"extras","doc":"```elixir\nplug :authorize_resource,\n  model: Post,\n  only: [:index, :new, :create],\n  required: false\n\nplug :load_and_authorize_resource,\n  model: Post,\n  except: [:index, :create, :new]\n```","title":"Example Usage - Getting Started","ref":"getting-started.html#example-usage"},{"type":"extras","doc":"If you need to load multiple resources for the `:index` action, you can either use a plug or load the resources directly within the `index/2` controller action.\n\n#### Option 1: Using a Plug\n```elixir\nplug :load_all_resources when action in [:index]\n\ndefp load_all_resources(conn, _opts) do\n  assign(conn, :posts, Posts.list_posts())\nend\n```\n\n#### Option 2: Loading Directly in the Controller Action\n```elixir\ndef index(conn, _params) do\n  posts = Posts.list_posts()\n  render(conn, \"index.html\", posts: posts)\nend\n```","title":"Loading All Resources in `:index` Action - Getting Started","ref":"getting-started.html#loading-all-resources-in-index-action"},{"type":"extras","doc":"Sometimes, you need to load and authorize a parent resource when dealing with nested relationships—such as when creating a child resource or listing all children of a parent. With the default `:required` set to true, if the parent resource is not found, the `not_found_handler` will be called.","title":"Nested Resources - Getting Started","ref":"getting-started.html#nested-resources"},{"type":"extras","doc":"When loading and authorizing a `Post` resource that `has_many` `Comment` resources:\n\n```elixir\n# Load and authorize the parent (Post)\nplug :load_and_authorize_resource,\n  model: Post,\n  id_name: \"post_id\",\n  only: [:create_comment]\n\n# Authorize action the child (Comment)\nplug :authorize_resource,\n  model: Comment,\n  only: [:create_comment, :save_comment],\n  required: false\n```\n\n#### Explanation\n\n1. The first plug loads and authorizes the parent `Post` resource using the `post_id` from `params` in the URL (`/posts/:post_id/comments`).\n   - The `:required` option ensures that if the Post is missing, the `not_found_handler` is called.\n2. The second plug authorizes actions on the child `Comment` resource.\n  - Since this is a **non-ID action**, `authorize_resource` is used.\n  - The `Comment` module name is passed as the resource to `can?/3` since no specific `Comment` does not exists yet.\n\nThis approach ensures that authorization is enforced correctly in nested resource scenarios.","title":"Example Usage - Getting Started","ref":"getting-started.html#example-usage"},{"type":"extras","doc":"To perform authorization checks, you need to implement the [`Canada.Can` protocol](https://github.com/jarednorman/canada) for each subject that requires permission validation.\nBy default, Canary uses `:current_user` from Plug or LiveView assigns as the subject.","title":"Defining Permissions - Getting Started","ref":"getting-started.html#defining-permissions"},{"type":"extras","doc":"Assume your application has a `User` module for authentication.\nYou can define permissions in `lib/abilities/user.ex`:\n\n```elixir\ndefimpl Canada.Can, for: User do\n  # Super admin can do everything\n  def can?(%User{role: \"superadmin\"}, _action, _resource), do: true\n\n  # Post owner can view and modify their own posts\n  def can?(%User{id: user_id}, action, %Post{user_id: user_id})\n    when action in [:show, :edit, :update], do: true\n\n  # Deny all other actions by default\n  def can?(%User{id: user_id}, _, _), do: false\nend\n```","title":"Example: Defining Permissions for an Authenticated User - Getting Started","ref":"getting-started.html#example-defining-permissions-for-an-authenticated-user"},{"type":"extras","doc":"If the subject (`:current_user` in assigns) is `nil`, and the authorization check is performed then `can/3` will be performed against `Atom`.\n\nFor anonymous users, define permissions, for example: `lib/abilities/anonymous.ex`:\n```elixir\ndefimpl Canada.Can, for: Atom do\n  # Allow anonymous users to register\n  def can?(nil, :new, User), do: true\n  def can?(nil, :create, User), do: true\n  def can?(nil, :confirm, User), do: true\n\n  # Allow anonymous users to create sessions\n  def can?(nil, :new, Session), do: true\n  def can?(nil, :create, Session), do: true\n\n  # Deny all other actions\n  def can?(_, _action, _model), do: false\nend\n```\n\nDefining permissions for `Atom` and `nil` subjects is optional.\nIf your application enforces authentication using a plug like `:require_authenticated_user` in the router pipeline, this may not be necessary.","title":"Handling Anonymous Users - Getting Started","ref":"getting-started.html#handling-anonymous-users"},{"type":"extras","doc":"","title":"Error handling - Getting Started","ref":"getting-started.html#error-handling"},{"type":"extras","doc":"By default, when subject is unauthorized to access an action, Canary sets `assigns.authorized` to `false`.\nHowever, you can configure a custom handler function to be called when authorization fails.\nCanary will pass the `Plug.Conn` or `Phoenix.LiveView.Socket` to the specified function, which should accept `conn` or `socket` as its only argument and return a `Plug.Conn` or tuple `{:halt, socket}`.\n\nThe error handler should implement the `Canary.ErrorHandler` behavior.\nRefer to the default implementation in `Canary.DefaultHandler`.\n\nFor example, to have Canary call `ErrorHandler.handle_unauthorized/1`:\n\n```elixir\nconfig :canary, error_handler: ErrorHandler\n```\n\n> #### LiveView Hook handlers\n>\n> In LiveView, the error handler should return `{:halt, socket}`.\n> For `handle_params`, it should also perform a redirect.","title":"Handling Unauthorized Actions - Getting Started","ref":"getting-started.html#handling-unauthorized-actions"},{"type":"extras","doc":"By default, when a resource is not found, Canary sets the resource in `assigns` to `nil`.\nSimilar to unauthorized action handling, you can configure a function that Canary will call when a resource is missing. This function will receive the `conn` (for Plugs) or `socket` (for LiveView).\n\n```elixir\nconfig :canary, error_handler: ErrorHandler\n```","title":"Handling Resource Not Found - Getting Started","ref":"getting-started.html#handling-resource-not-found"},{"type":"extras","doc":"You can specify custom handlers per action using `opts` in the `plug` or `mount_canary` call.\nThese handlers will override any globally configured error handlers.\n\n<!-- tabs-open -->","title":"Overriding Handlers Per Action - Getting Started","ref":"getting-started.html#overriding-handlers-per-action"},{"type":"extras","doc":"```elixir\nplug :load_and_authorize_resource Post,\n  unauthorized_handler: {Helpers, :handle_unauthorized},\n  not_found_handler: {Helpers, :handle_not_found}\n```\n\n> **Tip:** If you want to stop request handling after the handler function executes (e.g., for a redirect),\n> be sure to call `Plug.Conn.halt/1` within your handler:\n\n```elixir\ndef handle_unauthorized(conn) do\n  conn\n  |> put_flash(:error, \"You can't access that page!\")\n  |> redirect(to: \"/\")\n  |> halt()\nend\n```","title":"Conn Plugs - Getting Started","ref":"getting-started.html#conn-plugs"},{"type":"extras","doc":"```elixir\nmount_canary :load_and_authorize_resource Post,\n  unauthorized_handler: {Helpers, :handle_unauthorized},\n  not_found_handler: {Helpers, :handle_not_found}\n```\n\n> **Tip:** If you want to stop request handling after the handler function executes (e.g., for a redirect),\n> be sure to call `Plug.Conn.halt/1` within your handler:\n\n```elixir\ndef handle_unauthorized(socket) do\n  {:halt, Phoenix.LiveView.redirect(socket, to: \"/\")}\nend\n```\n<!-- tabs-close -->\n\n> #### Error handler order {: .info}\n>\n> If both an `:unauthorized_handler` and a `:not_found_handler` are specified for `load_and_authorize_resource`,\n> and the request meets the criteria for both, the `:unauthorized_handler` will be called first.","title":"LiveView Hooks - Getting Started","ref":"getting-started.html#liveview-hooks"},{"type":"extras","doc":"Canary\n======\n[![Actions Status](https://github.com/cpjk/canary/workflows/CI/badge.svg)](https://github.com/runhyve/canary/actions?query=workflow%3ACI)\n[![Hex pm](https://img.shields.io/hexpm/v/canary.svg?style=flat)](https://hex.pm/packages/canary)\n\nAn authorization library in Elixir for `Plug` and `Phoenix.LiveView` applications that restricts what resources the current user is allowed to access, and automatically load and assigns resources.\n\nInspired by [CanCan](https://github.com/CanCanCommunity/cancancan) for Ruby on Rails.\n\n[Read the docs](https://hexdocs.pm/canary/2.0.0-dev/getting-started.html)\n\n# Canary 2.0.0\n\nThe `master` branch is for the development of Canary 2.0.0. Check out [branch 1.2.x](https://github.com/cpjk/canary/tree/1.2.x) if you are looking Canary 1 (only plug authentication).","title":"Canary","ref":"readme.html"},{"type":"extras","doc":"For the latest master (2.0.0-dev):\n\n```elixir\ndefp deps do\n  {:canary, github: \"cpjk/canary\"}\nend\n```\n\nFor the latest release:\n\n```elixir\ndefp deps do\n  {:canary, \"~> 2.0.0-dev\"}\nend\n```\n\nThen run `mix deps.get` to fetch the dependencies.","title":"Installation - Canary","ref":"readme.html#installation"},{"type":"extras","doc":"Canary provides functions to be used as plugs or LiveView hooks to load and authorize resources:\n\n`load_resource`, `authorize_resource`, `authorize_controller`*, and `load_and_authorize_resource`.\n\n`load_resource` and `authorize_resource` can be used by themselves, while `load_and_authorize_resource` combines them both.\n\n*Available only in plug based authentication*\n\nIn order to use Canary, you will need, at minimum:\n\n- A [Canada.Can protocol](https://github.com/jarednorman/canada) implementation (a good place would be `lib/abilities.ex`)\n\n- An Ecto record struct containing the user to authorize in `assigns.current_user` (the key can be customized - [see more](#overriding-the-default-user)).\n\n- Your Ecto repo specified in your `config/config.exs`: `config :canary, repo: YourApp.Repo`\n\nFor the plugs just `import Canary.Plugs`. In a Phoenix app the best place would probably be inside `controller/0` in your `web/web.ex`, in order to make the functions available in all of your controllers.\n\nFor the liveview hooks just `use Canary.Hooks`. In a Phoenix app the best place would probably be inside `live_view/0` in your `web/web.ex`, in order to make the functions available in all of your controllers.","title":"Quick start - Canary","ref":"readme.html#quick-start"},{"type":"extras","doc":"Loads the resource having the id given in `params[\"id\"]` from the database using the given Ecto repo and model, and assigns the resource to `assigns. `, where `resource_name` is inferred from the model name.\n\n<!-- tabs-open -->","title":"load_resource - Canary","ref":"readme.html#load_resource"},{"type":"extras","doc":"```elixir\nplug :load_resource, model: Project.Post\n```\n\nWill load the `Project.Post` having the id given in `conn.params[\"id\"]` through `YourApp.Repo`, and assign it to `conn.assigns.post`.","title":"Conn Plugs example - Canary","ref":"readme.html#conn-plugs-example"},{"type":"extras","doc":"```elixir\nmount_canary :load_resource, model: Project.Post\n```\n\nWill load the `Project.Post` having the id given in `params[\"id\"]` through `YourApp.Repo`, and assign it to `socket.assigns.post`\n<!-- tabs-close -->","title":"LiveView Hooks example - Canary","ref":"readme.html#liveview-hooks-example"},{"type":"extras","doc":"Checks whether or not the `current_user` for the request can perform the given action on the given resource and assigns the result (true/false) to `assigns.authorized`. It is up to you to decide what to do with the result.\n\nFor Phoenix applications, Canary determines the action automatically.\nFor non-Phoenix applications, or to override the action provided by Phoenix, simply ensure that `assigns.canary_action` contains an atom specifying the action.\n\nFor the LiveView on `handle_params` it uses `socket.assigns.live_action` as action, on `handle_event` it uses the event name as action.\n\n\n\nIn order to authorize resources, you must specify permissions by implementing the [Canada.Can protocol](https://github.com/jarednorman/canada) for your `User` model (Canada is included as a light weight dependency).","title":"authorize_resource - Canary","ref":"readme.html#authorize_resource"},{"type":"extras","doc":"Authorizes the resource and then loads it if authorization succeeds. Again, the resource is loaded into `assigns. `.\n\nIn the following example, the `Post` with the same `user_id` as the `current_user` is only loaded if authorization succeeds.","title":"load_and_authorize_resource - Canary","ref":"readme.html#load_and_authorize_resource"},{"type":"extras","doc":"Let's say you have a Phoenix application with a `Post` model, and you want to authorize the `current_user` for accessing `Post` resources.\n\nLet's suppose that you have a file named `lib/abilities.ex` that contains your Canada authorization rules like so:\n\n```elixir\ndefimpl Canada.Can, for: User do\n  def can?(%User{ id: user_id }, action, %Post{ user_id: user_id })\n    when action in [:show], do: true\n\n  def can?(%User{ id: user_id }, _, _), do: false\nend\n```","title":"Usage Example - Canary","ref":"readme.html#usage-example"},{"type":"extras","doc":"In your `web/router.ex:` you have:\n\n```elixir\nget \"/posts/:id\", PostController, :show\ndelete \"/posts/:id\", PostController, :delete\n```\n\nTo automatically load and authorize on the `Post` having the `id` given in the params, you would add the following plug to your `PostController`:\n\n```elixir\nplug :load_and_authorize_resource, model: Post\n```\n\nIn this case, on `GET /posts/12` authorization succeeds, and the `Post` specified by `conn.params[\"id]` will be loaded into `conn.assigns.post`.\n\nHowever, on `DELETE /posts/12`, authorization fails and the `Post` resource is not loaded.","title":"Example for Conn Plugs - Canary","ref":"readme.html#example-for-conn-plugs"},{"type":"extras","doc":"In your `web/router.ex:` you have:\n\n```elixir\nlive \"/posts/:id\", PostLive, :show\n```\n\nand in your PostLive module `web/live/post_live.ex`:\n\n```elixir\ndefmodule MyAppWeb.PostLive do\n  use MyAppWeb, :live_view\n\n  def render(assigns) do\n    ~H\"\"\"\n    Post id: {@post.id}\n     Delete \n    \"\"\"\n  end\n\n  def mount(_params, _session, socket), do: {:ok, socket}\n\n  def handle_event(\"delete\", _params, socket) do\n    # Do the action\n    {:noreply, update(socket, :temperature, &(&1 + 1))}\n  end\nend\n```\n\nTo automatically load and authorize on the `Post` having the `id` given in the params, you would add the following hook to your `PostLive`:\n\n```elixir\nmount_hook :load_and_authorize_resource, model: Post\n```\n\nIn this case, once opening `/posts/12` the `load_and_authorize_resource` on `handle_params` stage will be performed. The the `Post` specified by `params[\"id]` will be loaded into `socket.assigns.post`.\n\nHowever, when the `delete` event will be triggered, authorization fails and the `Post` resource is not loaded. Socket will be halted.","title":"Example for LiveView Hooks - Canary","ref":"readme.html#example-for-liveview-hooks"},{"type":"extras","doc":"To exclude an action from any of the plugs, pass the `:except` key, with a single action or list of actions.\n\nFor example,\n\nSingle action form:\n\n```elixir\nplug :load_and_authorize_resource, model: Post, except: :show\n\nmount_canary :load_and_authorize_resource, model: Post, except: :show\n```\n\nList form:\n\n```elixir\nplug :load_and_authorize_resource, model: Post, except: [:show, :create]\n\nmount_canary :load_and_authorize_resource, model: Post, except: [:show, :create]\n```","title":"Excluding actions - Canary","ref":"readme.html#excluding-actions"},{"type":"extras","doc":"To specify that a plug should be run only for a specific list of actions, pass the `:only` key, with a single action or list of actions.\n\nFor example,\n\nSingle action form:\n\n```elixir\nplug :load_and_authorize_resource, model: Post, only: :show\n\nmount_canary :load_and_authorize_resource, model: Post, only: :show\n```\n\nList form:\n\n```elixir\nplug :load_and_authorize_resource, model: Post, only: [:show, :create]\n\nmount_canary :load_and_authorize_resource, model: Post, only: [:show, :create]\n```\n\n> Note: Having both `:only` and `:except` in opts is invalid. Canary will raise `ArgumentError` \"You can't use both :except and :only options\"","title":"Authorizing only specific actions - Canary","ref":"readme.html#authorizing-only-specific-actions"},{"type":"extras","doc":"Globally, the default key for finding the user to authorize can be set in your configuration as follows:\n\n```elixir\nconfig :canary, current_user: :some_current_user\n```\n\nIn this case, canary will look for the current user record in `assigns.some_current_user`.\n\nThe current user key can also be overridden for individual plugs as follows:\n\n```elixir\nplug :load_and_authorize_resource, model: Post, current_user: :current_admin\n\nmount_canary :load_and_authorize_resource, model: Post, current_user: :current_admin\n```","title":"Overriding the default user - Canary","ref":"readme.html#overriding-the-default-user"},{"type":"extras","doc":"To specify the name under which the loaded resource is stored, pass the `:as` flag in the plug declaration.\n\nFor example,\n\n```elixir\nplug :load_and_authorize_resource, model: Post, as: :new_post\n\nmount_canary :load_and_authorize_resource, model: Post, as: :new_post\n```\n\nwill load the post into `assigns.new_post`","title":"Specifying resource_name - Canary","ref":"readme.html#specifying-resource_name"},{"type":"extras","doc":"Associations can be preloaded with `Repo.preload` by passing the `:preload` option with the name of the association:\n\n```elixir\nplug :load_and_authorize_resource, model: Post, preload: :comments\n\nmount_canary :load_and_authorize_resource, model: Post, preload: :comments\n```","title":"Preloading associations - Canary","ref":"readme.html#preloading-associations"},{"type":"extras","doc":"To authorize actions where there is no loaded resource, the resource passed to the `Canada.Can` implementation should be the module name of the model rather than a struct.\n\nTo authorize such actions use `authorize_resource` plug with `required: false` option\n\n```elixir\nplug :authorize_resource, model: Post, only: [:index, :new, :create], required: false\n\nmount_canary :authorize_resource, model: Post, only: [:index, :new, :create], required: false\n```\n\nFor example, when authorizing access to the `Post` resource, you should use\n\n```elixir\ndef can?(%User{}, :index, Post), do: true\n```\n\ninstead of\n\n```elixir\ndef can?(%User{}, :index, %Post{}), do: true\n```\n\n>","title":"Non-id actions - Canary","ref":"readme.html#non-id-actions"},{"type":"extras","doc":">\n> The `:non_id_actions` is deprecated as of 2.0.0-dev and will be removed in Canary 2.1.0\n> Please follow the [Upgrade guide to 2.0.0](docs/upgrade.md#upgrading-from-canary-1-2-0-to-2-0-0) for more details.","title":"Deprecated {: .warning} - Canary","ref":"readme.html#deprecated-warning"},{"type":"extras","doc":"Sometimes you need to load and authorize a parent resource when you have\na relationship between two resources and you are creating a new one or\nlisting all the children of that parent. Depending on your authorization\nmodel you migth authorize against the parent resource or against the child.\n\n```elixir\ndefmodule MyAppWeb.CommentController do\n\n  plug :load_and_authorize_resource,\n    model: Post,\n    id_name: \"post_id\",\n    only: [:new_comment, :create_comment]\n\n  # get /posts/:post_id/comments/new\n  def new_comment(conn, _params) do\n    # ...\n  end\n\n  # post /posts/:post_id/comments\n  def new_comment(conn, _params) do\n    # ...\n  end\nend\n```\n\nIt will authorize using `Canada.Can` with following arguments:\n1. subject is `conn.assigns.current_user`\n2. action is `:new_comment` or `:create_comment`\n3. resource is `%Post{}` with `conn.params[\"post_id\"]`\n\nThanks to the `:requried` set to true by default this plug will call `not_found_handler` if the `Post` with given `post_id` does not exists.\nIf for some reason you want to disable it, set `required: false` in opts.\n\n>","title":"Nested associations - Canary","ref":"readme.html#nested-associations"},{"type":"extras","doc":">\n> The `:persisted` is deprecated as of 2.0.0-dev and will be removed in Canary 2.1.0\n> Please follow the [Upgrade guide to 2.0.0](docs/upgrade.md#upgrading-from-canary-1-2-0-to-2-0-0) for more details.","title":"Deprecated {: .warning} - Canary","ref":"readme.html#deprecated-warning"},{"type":"extras","doc":"You may wish to define permissions for when there is no logged in current user (when `conn.assigns.current_user` is `nil`).\nIn this case, you should implement `Canada.Can` for `nil` like so:\n\n```elixir\ndefimpl Canada.Can, for: Atom do\n  # When the user is not logged in, all they can do is read Posts\n  def can?(nil, :show, %Post{}), do: true\n  def can?(nil, _, _), do: false\nend\n```","title":"Implementing Canada.Can for an anonymous user - Canary","ref":"readme.html#implementing-canada-can-for-an-anonymous-user"},{"type":"extras","doc":"You can tell Canary to search for a resource using a field other than the default `:id` by using the `:id_field` option. Note that the specified field must be able to uniquely identify any resource in the specified table.\n\nFor example, if you want to access your posts using a string field called `slug`, you can use\n\n```elixir\nplug :load_and_authorize_resource, model: Post, id_name: \"slug\", id_field: \"slug\"\n```\n\nto load and authorize the resource `Post` with the slug specified by `conn.params[\"slug\"]` value.\n\nIf you are using Phoenix, your `web/router.ex` should contain something like:\n\n```elixir\nresources \"/posts\", PostController, param: \"slug\"\n```\n\nThen your URLs will look like:\n\n```\n/posts/my-new-post\n```\n\ninstead of\n\n```\n/posts/1\n```","title":"Specifing database field - Canary","ref":"readme.html#specifing-database-field"},{"type":"extras","doc":"By default, when an action is unauthorized, Canary simply sets `conn.assigns.authorized` to `false`.\nHowever, you can configure a handler function to be called when authorization fails. Canary will pass the `Plug.Conn` to the given function. The handler should accept a `Plug.Conn` as its only argument, and should return a `Plug.Conn`.\n\nFor example, to have Canary call `Helpers.handle_unauthorized/1`:\n\n```elixir\nconfig :canary, unauthorized_handler: {Helpers, :handle_unauthorized}\n```","title":"Handling unauthorized actions - Canary","ref":"readme.html#handling-unauthorized-actions"},{"type":"extras","doc":"By default, when a resource is not found, Canary simply sets the resource in `conn.assigns` to `nil`. Like unauthorized action handling , you can configure a function to which Canary will pass the `conn` when a resource is not found:\n\n```elixir\nconfig :canary, not_found_handler: {Helpers, :handle_not_found}\n```\n\nYou can also specify handlers on an individual basis (which will override the corresponding configured handler, if any) by specifying the corresponding `opt` in the plug call:\n\n```elixir\nplug :load_and_authorize_resource Post,\n  unauthorized_handler: {Helpers, :handle_unauthorized},\n  not_found_handler: {Helpers, :handle_not_found}\n```\n\nTip: If you would like the request handling to stop after the handler function exits, e.g. when redirecting, be sure to call `Plug.Conn.halt/1` within your handler like so:\n\n```elixir\ndef handle_unauthorized(conn) do\n  conn\n  |> put_flash(:error, \"You can't access that page!\")\n  |> redirect(to: \"/\")\n  |> halt\nend\n```\n\nNote: If both an `:unauthorized_handler` and a `:not_found_handler` are specified for `load_and_authorize_resource`, and the request meets the criteria for both, the `:unauthorized_handler` will be called first.","title":"Handling resource not found - Canary","ref":"readme.html#handling-resource-not-found"},{"type":"extras","doc":"MIT License. Copyright 2016 Chris Kelly.","title":"License - Canary","ref":"readme.html#license"},{"type":"extras","doc":"","title":"CHANGELOG","ref":"changelog.html"},{"type":"extras","doc":"","title":"Changelog - CHANGELOG","ref":"changelog.html#changelog"},{"type":"extras","doc":"Canary 2.0.0 introduces authorization hooks for Phoenix LiveView. The Plug based authorization was refactored a bit to make the API cosistent. Please follow the [Upgrade guide to 2.0.0](docs/upgrade.md#upgrading-from-canary-1-2-0-to-2-0-0) for more details.\n\n  * Enhancements\n    * added support for authorization LiveView with `Canary.Hooks`\n    * added `:error_handler` and ErrorHandler behaviour\n    * added `:required` option, default to true\n\n  * Dependency changes\n    * Elixir ~> 1.14 is now required\n\n  * Deprecations\n    * The `:non_id_actions` option is deprecated and will be removed in Canary 2.1.0. Use separate `:authorize_resource` plug for `non_id_actions` and `:except` to exclude non_in_actions.\n    * The `:persisted` option is deprecated and will be removed in Canary 2.1.0. Use `:required` instead.","title":"v2.0.0-dev - CHANGELOG","ref":"changelog.html#v2-0-0-dev"},{"type":"extras","doc":"* Enhancements\n    * Add `required` opt","title":"v1.2.0 - CHANGELOG","ref":"changelog.html#v1-2-0"},{"type":"extras","doc":"* Enhancements\n    * Add `non_id_actions` opt","title":"v1.1.0 - CHANGELOG","ref":"changelog.html#v1-1-0"},{"type":"extras","doc":"* Bug fixes\n    * Do not clobber resources in the `Conn` on index action if they are of the same model","title":"v1.0 - CHANGELOG","ref":"changelog.html#v1-0"},{"type":"extras","doc":"* Relax Ecto version requirements","title":"v0.14.2 - CHANGELOG","ref":"changelog.html#v0-14-2"},{"type":"extras","doc":"* Bug fixes\n    * Use Macro.underscore/1 instead of Mix.Utils.underscore/1 to avoid :mix dependency on production","title":"v0.14.1 - CHANGELOG","ref":"changelog.html#v0-14-1"},{"type":"extras","doc":"* Enhancements\n    * You can now tell Canary to search for a resource using a field other than the default `:id` by using the `:id_field` option. Note that the specified field must be able to uniquely identify any resource in the specified table.\n  * Dependency changes\n    * Elixir ~> 1.2 is now required\n    * Ecto ~> 1.1 is now required","title":"v0.14.0 - CHANGELOG","ref":"changelog.html#v0-14-0"},{"type":"extras","doc":"* Enhancements\n    * If both an `:unauthorized_handler` and a `:not_found_handler` are specified for `load_and_authorize_resource`, and the request meets the criteria for both, the `:unauthorized_handler` will be called first.\n  * Bug Fixes\n    * If more than one handler are specified and the first handler halts the request, the second handler will be skipped.","title":"v0.13.1 - CHANGELOG","ref":"changelog.html#v0-13-1"},{"type":"extras","doc":"* Enhancements\n    * Canary can now be configured to call a user-defined function when a resource is not found. The function is specified and used in a similar manner to `:unauthorized_handler`.\n  * Bug Fixes\n    * Disabled protocol consolidation in order for tests to work on Elixir 1.2","title":"v0.13.0 - CHANGELOG","ref":"changelog.html#v0-13-0"},{"type":"extras","doc":"* Deprecations\n    * Canary now looks for the current action in `conn.assigns.canary_action` rather than `conn.assigns.action` in order to avoid conflicts. The `action` key is deprecated.","title":"v0.12.2 - CHANGELOG","ref":"changelog.html#v0-12-2"},{"type":"extras","doc":"* Enhancements\n  * Canary can now be configured to call a user-defined function when authorization fails. Canary will pass the `Plug.Conn` for the request to the given function. The handler should accept a `Plug.Conn` as its only argument, and should return a `Plug.Conn`.\n    * For example, to have Canary call `Helpers.handle_unauthorized/1`:\n    ```elixir\n    config :canary, unauthorized_handler: {Helpers, :handle_unauthorized}\n    ```\n    * You can also specify the `:unauthorized_handler` on an individual basis by specifying the `:unauthorized_handler`   `opt` in the plug call like so:\n    ```elixir\n    plug :load_and_authorize_resource Post, unauthorized_handler: {Helpers, :handle_unauthorized}\n    ```","title":"v0.12.0 - CHANGELOG","ref":"changelog.html#v0-12-0"},{"type":"extras","doc":"* Enhancements\n  * Resources can now be loaded on `:new` and `:create` actions, when `persisted: true` is specified in the plug call. This allows parent resources to be loaded when a child is created. For example, if a `Post` resource has multiple `Comment` children, you may want to load the parent `Post` when creating a new `Comment`. You can load the parent `Post` with a separate\n  ```elixir\n  plug :load_and_authorize_resource, model: Post, id_name: \"post_id\", persisted: true, only: [:create]\n  ```\n  This will cause Canary to try to load the corresponding `Post` from the database when creating a `Comment` at the URL `/posts/:post_id/comments`","title":"v0.11.0 - CHANGELOG","ref":"changelog.html#v0-11-0"},{"type":"extras","doc":"* Bug fix\n  * Correctly checks `conn.assigns` for pre-existing resource\n\n* Deprecations\n  * Canary now favours looking for the current action in `conn.assigns.canary_action` rather than `conn.assigns.action` in order to avoid conflicts. The `action` key is deprecated\n\n* Enhancements\n  * The name of the id in `conn.params` can now be specified with the `id_name` opt","title":"v0.10.0 - CHANGELOG","ref":"changelog.html#v0-10-0"},{"type":"extras","doc":"# Upgrade guides","title":"Upgrade guides","ref":"upgrade.html"},{"type":"extras","doc":"","title":"Upgrading from Canary 1.2.0 to 2.0.0 - Upgrade guides","ref":"upgrade.html#upgrading-from-canary-1-2-0-to-2-0-0"},{"type":"extras","doc":"> Since 2.0.0, the `:persisted` and `:non_id_actions` options have been deprecated and will be removed in Canary 2.1.0.\n\nYou need to update plug calls. Using `:authorize_resource` for actions where there is no actual load action is more explicit.\n\nLet's assume you currently have the following plug call:\n```elixir\n  plug :load_and_authorize_resource,\n    model: Network,\n    non_id_actions: [:index, :create, :new],\n    preload: [:hypervisor]\n```\n\nNow let's break it apart:\n```elixir\n  plug :authorize_resource,\n    model: Network,\n    only: [:index, :create, :new],\n    required: false\n\n  plug :load_and_authorize_resource,\n    model: Network,\n    except: [:index, :create, :new],\n    preload: [:hypervisor]\n```\n\nFor non-ID actions, there is a separate plug for authorization. The `required: false` option marks the resource as optional during authorization checks, and the model module name is used. Essentially, this is how :non_id_actions worked.\nFor actions other than `:index`, `:create`, and `:new`, it will load and authorize resources as usual.\nTo load all resources in the `:index` action, you can set up a plug or add the load function directly in `index/2`.\n\n```elixir\n  # using a plug\n\n  plug :load_all_resources when action in [:index]\n  defp load_all_resources(conn, _opts) do\n    assign(:networks, Hypervisors.list_hypervisor_networks(hypervisor))\n  end\n\n  # directly in the controller action\n\n  def index(conn, _params) do\n    networks = Hypervisors.list_hypervisor_networks(hypervisor)\n    render(conn, \"index.html\", networks: networks)\n  end\n```","title":"Update Your Non-ID Actions - Upgrade guides","ref":"upgrade.html#update-your-non-id-actions"},{"type":"extras","doc":"With the [update non-id action](#update-your-non-id-actions) the `:persisted` is no longer required.","title":"Remove `:persisted` option - Upgrade guides","ref":"upgrade.html#remove-persisted-option"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.0"}}